"""Mixin module.

This module contains the main VIPickable class and its meta class.
"""
import pickle
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional, Union

import orjson
from loguru import logger

from .errors import DumpAttributeError, RestoreAttributeError
from .save_utils import create_folder

DUMP_METHOD_PATTERN = "_dump_{}_"
RESTORE_METHOD_PATTERN = "_restore_{}_"


class MetaVIPicklable(type):
    """Metaclass for VIPicklable.

    This metaclass is aimed to be used with VIPicklable class. It add the functionality
    to inherit attributes from the parent classes. Which is usefull for adding or
    removing attributes from PICKLE_BLACKLIST and CONFIG_ITEMS.
    """

    def __new__(cls, name: str, parents: tuple, attributes: dict):
        """Metaclass constructor.

        Construct CONFIG_ITEMS and PICKLE_BLACKLIST.

        Args:
            name (str): Name of the class
            parents (tuple): Parent classes
            attributes (dict): Class attributes
        """
        for prefix in ("PICKLE_BLACKLIST", "CONFIG_ITEMS"):
            # If the class has a prefix attribute, it will be used as it is
            # we convert it to a set to prevent duplicated values
            if prefix in attributes:
                pickle_blacklist = {key for key in attributes[prefix]}

            # Otherwise the prefix attribute will be created from the parent classes
            # prefix attributes and the {prefix}_ADD and {prefix}_REMOVE attributes
            else:
                pickle_blacklist = set()

                # First we get the union of all pickle_blacklist of the parents
                for parent in parents:
                    if issubclass(parent, VIPicklable):
                        pickle_blacklist = pickle_blacklist.union(
                            getattr(parent, prefix, set())
                        )

                # Then we add attributes from {prefix}_ADD
                pickle_blacklist = pickle_blacklist.union(
                    attributes.get(f"{prefix}_ADD", set())
                )
                # and we remove attributes from {prefix}_REMOVE
                pickle_blacklist = pickle_blacklist.difference(
                    attributes.get(f"{prefix}_REMOVE", set())
                )

            # Finnaly, {prefix} is converted to a tuple for immutablity
            attributes[prefix] = tuple(sorted(pickle_blacklist))

        return super().__new__(cls, name, parents, attributes)


class VIPicklable(metaclass=MetaVIPicklable):
    """VIPicklable main class.

    Inherit from this class and define attributes to blacklist and the ones to add to
    the object configuration dict.
    """

    PICKLE_NAME: str = "archive.pkl"
    PICKLE_BLACKLIST: Iterable[str] = ()
    PICKLE_BLACKLIST_ADD: Iterable[str] = ()
    PICKLE_BLACKLIST_REMOVE: Iterable[str] = ()

    CONFIG_NAME: str = "config.json"
    CONFIG_ITEMS: Iterable[str] = ()
    CONFIG_ITEMS_ADD: Iterable[str] = ()
    CONFIG_ITEMS_REMOVE: Iterable[str] = ()

    @property
    def configurations(self) -> dict:
        """A configuration dict used to show important attributes.

        Returns:
            dict: configuration dict
        """
        return {
            key: getattr(self, key) for key in self.CONFIG_ITEMS if hasattr(self, key)
        }

    def __getstate__(self):
        """Pickle all attributes except the ones listed in PICKLE_BLACKLIST."""
        return {
            attribute: state
            for attribute, state in self.__dict__.items()
            if attribute not in self.PICKLE_BLACKLIST
        }

    def save(
        self,
        path: Union[str, Path],
        pickle_dump_kwargs: dict = None,
        json_dump_kwargs: dict = None,
        overwrite: bool = True,
    ):
        """Save the object instance in a dedicated directory.

        Args:
            path (Union[str, Path]): Path to the directory
            pickle_dump_kwargs (dict, optional): kwargs to be passed to
                pickle.dump method. Defaults to None.
            json_dump_kwargs (dict, optional): kwargs to be passed to json.dump method.
                Defaults to None.
            overwrite (bool, optional): If True, overwrite the folder if it exists.
        """
        # Before save hook
        self.before_save()

        path = create_folder(path)

        if pickle_dump_kwargs is None:
            pickle_dump_kwargs = {}

        if json_dump_kwargs is None:
            json_dump_kwargs = {}

        self.save_instance(path, overwrite=overwrite, **json_dump_kwargs)
        self.save_config(path, overwrite=overwrite, **json_dump_kwargs)
        self.save_pickle_blacklisted(path, overwrite=overwrite)

        # After save hook
        self.after_save()

    def before_save(self):
        """Hook executed at the beggining of the save method."""

    def after_save(self):
        """Hook executed at the end of the save method."""

    def save_instance(self, path: Union[str, Path], overwrite: bool = True, **kwargs):
        """Save the current instance.

        Args:
            path (Union[str, Path]): path to a folder where to save the current instance
            overwrite (bool, optional): If True, overwrite the folder if it exists.
            kwargs (dict, optional): Additional parameters. Default to {}.
        """
        if not self.PICKLE_NAME:
            logger.info(
                f"{self.__class__.__name__}.PICKLE_NAME is empty so the instance "
                f"will not be saved"
            )
            return

        path = create_folder(path)
        filepath = path / self.PICKLE_NAME

        if overwrite or not filepath.exists():
            with open(filepath, "wb") as f:
                pickle.dump(self, f, **kwargs)

    def save_config(
        self,
        path: Union[str, Path],
        overwrite: bool = True,
        option: Optional[int] = orjson.OPT_INDENT_2,
        default: Optional[Callable[[Any], Any]] = ...,
    ):
        """Save the instance configuration attributes.

        Args:
            path (Union[str, Path]): path to a folder where to save the current instance
                config file
            overwrite (bool, optional): If True, overwrite the folder if it exists.
            option (int, optional): ORJson options.
                See [orjson documentation](https://github.com/ijl/orjson).
                Defaults to orjson.OPT_INDENT_2.
            default (Callable): To serialize a subclass or arbitrary types, specify
                default as a callable that returns a supported type. default may be a
                function, lambda, or callable class instance. To specify that a type was
                not handled by default, raise an exception such as TypeError.
        """
        if not self.CONFIG_NAME:
            logger.info(
                f"{self.__class__.__name__}.CONFIG_NAME is empty so the configuration "
                f"dict will not be saved"
            )
            return

        path = create_folder(path)
        filepath = path / self.CONFIG_NAME

        if overwrite or not filepath.exists():
            with open(filepath, "wb") as f:
                f.write(
                    orjson.dumps(self.configurations, default=default, option=option)
                )

    def save_pickle_blacklisted(
        self, path: Union[str, Path], overwrite: bool = True
    ) -> Dict[str, Exception]:
        """Try to save excluded attributes.

        Args:
            path (Union[str, Path]): path to a folder where to save blacklisted
                attributes
            overwrite (bool, optional): If True, overwrite the folder if it exists
        Returns:
            Dict[str, Exception]: A dictionary of attributes that could not be saved
                and corresponding exceptions raised
        """
        path = create_folder(path)
        failures = {}

        for attribute in self.PICKLE_BLACKLIST:
            try:
                getattr(self, DUMP_METHOD_PATTERN.format(attribute))(
                    path, overwrite=overwrite
                )
            except DumpAttributeError as e:
                logger.warning(
                    f"{self.__class__.__name__}"
                    f".{DUMP_METHOD_PATTERN.format(attribute)} failed : "
                    f"self.{attribute} could not be dumped"
                )
                logger.exception(e)
                failures[attribute] = e
            except AttributeError as e:
                logger.debug(
                    f"self.{attribute} count not be dumped since there is no "
                    f"method {self.__class__.__name__}"
                    f".{DUMP_METHOD_PATTERN.format(attribute)}"
                )
                failures[attribute] = e

        return failures

    @classmethod
    def load_instance(cls, path: Union[str, Path], **kwargs) -> "VIPicklable":
        """Load a VIPicklable instance.

        Load an instance from a file or folder without unpicklable attributes.

        Args:
            path (Union[str, Path]): Path to the pickle file
            kwargs (dict, optional): Additional parameters. Default to {}.

        Raises:
            FileNotFoundError: Pickle file not found
        Returns:
            VIPicklable: The instance object
        """
        with open(path, "rb") as f:
            return pickle.load(f, **kwargs)

    @classmethod
    def load(
        cls,
        path: Union[str, Path],
        pickle_dump_kwargs: dict = None,
    ) -> "VIPicklable":
        """Load a VIPicklable instance.

        Load a VIPicklable instance and all loadable attributes from a file or folder.

        Args:
            path (Union[str, Path]): Path to the pickle file
            pickle_dump_kwargs (dict, optional): additionnal arguments to be pass to
                load_instance method. Default to None.

        Raises:
            FileNotFoundError: Pickle file not found
        Returns:
            VIPicklable: The instance object
        """
        if isinstance(path, str):
            path = Path(path)

        if pickle_dump_kwargs is None:
            pickle_dump_kwargs = {}

        if path.is_dir():
            pickle_path = path / cls.PICKLE_NAME
            folder_path = path
        elif path.is_file():
            pickle_path = path
            folder_path = path.parent
        else:
            raise FileNotFoundError(f"{path} not found")

        cls.before_load()

        obj = cls.load_instance(pickle_path, **pickle_dump_kwargs)
        obj.load_pickle_blacklisted(folder_path)

        obj.after_load()
        return obj

    @classmethod
    def before_load(cls):
        """Hook executed at the beggining of the load method."""

    @classmethod
    def after_load(cls):
        """Hook executed at the end of the load method."""

    def load_pickle_blacklisted(self, path: Union[str, Path]) -> Dict[str, Exception]:
        """Try to unpickle excluded attributes.

        Args:
            path (Union[str, Path]): Path to the pickle file

        Returns:
            Dict[str, Exception]: A dictionary of attributes that could not be loaded
                and corresponding exceptions raised
        """
        failures = {}
        for attribute in self.PICKLE_BLACKLIST:
            try:
                getattr(self, RESTORE_METHOD_PATTERN.format(attribute))(path)
            except RestoreAttributeError as e:
                logger.warning(
                    f"{self.__class__.__name__}"
                    f".{RESTORE_METHOD_PATTERN.format(attribute)} failed : "
                    f"self.{attribute} could not be restored"
                )
                logger.exception(e)
                failures[attribute] = e
            except AttributeError as e:
                logger.debug(
                    f"self.{attribute} count not be unpickled since there is no "
                    f"method {self.__class__.__name__}"
                    f".{RESTORE_METHOD_PATTERN.format(attribute)}"
                )
                failures[attribute] = e

        return failures



"""vipckle.save_utils tests."""

from pathlib import Path

import pytest
from vipickle import save_utils


def test_create_folder(tmp_path: Path):
    """Test create_folder function."""
    created_folder = save_utils.create_folder(tmp_path / "a/b/c")
    assert created_folder.is_dir()

    created_folder = save_utils.create_folder(f"{tmp_path}/1/2")
    assert created_folder.is_dir()

    with pytest.raises(FileExistsError):
        created_folder = save_utils.create_folder(tmp_path / "a/b/c", exist_ok=False)

    with pytest.raises(FileNotFoundError):
        created_folder = save_utils.create_folder(tmp_path / "sub/sub", parents=False)



"""VIPickle.

VIPickle make any object picklable by specifying which attribute to blacklist when
pickling and how to dump and recover them.
"""

from .errors import DumpAttributeError, RestoreAttributeError
from .mixin import MetaVIPicklable, VIPicklable
from .save_utils import create_folder

__all__ = [
    "DumpAttributeError",
    "RestoreAttributeError",
    "VIPicklable",
    "MetaVIPicklable",
    "create_folder",
]



"""vipckle.mixin tests."""
import json
import pickle
import random
from pathlib import Path

import pytest
from vipickle import mixin


class CustomVIPicklableA(mixin.VIPicklable):
    """Custom VIPicklable object."""

    PICKLE_BLACKLIST = ["unpickable", "wont_recover"]

    def __init__(self, param1: int):
        """Initialization.

        Args:
            param1 (int): a parameter
        """
        self.param1 = param1
        self.unpickable = lambda x: x * self.param1
        self.wont_recover = "do not dump"

    def _restore_unpickable_(self, path: Path):
        self.unpickable = lambda x: x * self.param1


def test_pickle_CustomVIPicklableA(tmp_path: Path):
    """Test to save an instance of CustomVIPicklableA."""
    # Creating a random CustomVIPicklableA instance
    param1 = random.randint(1, 100)
    a = CustomVIPicklableA(param1)
    assert a.wont_recover

    # Saving it to a temp folder
    a.save(tmp_path)

    # By default, the instance is saved under tmp_path/customVIPicklablea.pkl
    a_path = tmp_path / CustomVIPicklableA.PICKLE_NAME
    a_config_path = tmp_path / CustomVIPicklableA.CONFIG_NAME

    assert a_config_path.is_file()

    # If we reload only the instance, the unpickable is not here
    with open(a_path, "rb") as f:
        restored_a = pickle.load(f)

    with pytest.raises(AttributeError):
        restored_a.unpickable(10)

    # But if we reload it completly it is reconstructed by _restore_unpickable_
    restored_a = CustomVIPicklableA.load(tmp_path)
    assert restored_a.unpickable(10) == 10 * param1

    with pytest.raises(AttributeError):
        restored_a.wont_recover

    # load also work with tmp_path/customVIPicklablea.pkl
    restored_a = CustomVIPicklableA.load(a_path)
    assert restored_a.unpickable(10) == 10 * param1

    # load also work with tmp_path/customVIPicklablea.
    with pytest.raises(FileNotFoundError):
        CustomVIPicklableA.load(tmp_path / "b.pkl")


class CustomVIPicklableB(CustomVIPicklableA):
    """Subclass of CustomVIPicklableA."""

    PICKLE_BLACKLIST_ADD = ["param3"]
    PICKLE_BLACKLIST_REMOVE = ["wont_recover"]

    def __init__(self, param1: int, param2: int, param3: int):
        """Initialization."""
        super(CustomVIPicklableB, self).__init__(param1)
        self.param2 = param2
        self.param3 = param3
        self.unpickable = lambda x: self.param1 * self.param2

    def _restore_unpickable_(self, path: Path):
        self.unpickable = lambda x: self.param1 * self.param2


def test_pickle_CustomVIPicklableB(tmp_path: Path):
    """Test VIPicklable inheritance."""
    # Creating a random CustomVIPicklableA instance
    param1 = random.randint(1, 100)
    param2 = random.randint(1, 100)
    param3 = random.randint(1, 100)
    b = CustomVIPicklableB(param1, param2, param3)
    assert b.wont_recover
    assert b.param1 == param1
    assert b.param2 == param2
    assert b.param3 == param3

    # Saving it to a temp folder
    b.save(tmp_path)

    # We reload it with CustomVIPicklableA load function, it will work since it is
    # calling load_instance, load_config and load_pickle_blacklisted internally
    restored_b = CustomVIPicklableA.load(tmp_path)

    assert restored_b.unpickable(10) == param1 * param2
    assert restored_b.wont_recover

    with pytest.raises(AttributeError):
        restored_b.param3


class CustomVIPicklableC(CustomVIPicklableB):
    """Subclass of CustomVIPicklableA."""

    PICKLE_NAME = None
    CONFIG_NAME = ""
    PICKLE_BLACKLIST_ADD = ["wont_recover"]

    def _dump_wont_recover_(self, path: Path, overwrite: bool = False):
        with open(path / "wont_recover.json", "w") as f:
            json.dump({"wont_recover": "or maybe not"}, f)


def test_pickle_CustomVIPicklableC(tmp_path: Path):
    """Test an object that is not saved as a pickle."""
    c = CustomVIPicklableC(1, 2, 3)
    c.save(tmp_path)

    assert not list(tmp_path.glob("*.pkl"))
    assert (tmp_path / "wont_recover.json").exists()



"""VIPickle errors.

VIPickle custom errors
"""


class DumpAttributeError(AttributeError):
    """Use to indicate that the dumping method is not implemented."""


class RestoreAttributeError(AttributeError):
    """Use to indicate that the restoring method is not implemented."""



"""Saving utilities module.

Module containing utilities functions for saving purpose.
"""
from pathlib import Path
from typing import Union


def create_folder(
    path: Union[str, Path], exist_ok: bool = True, parents: bool = True
) -> Path:
    """Create a folder if it does not exists and returns it.

    Args:
        path (Union[str, Path]): Path to the folder
        exist_ok (bool, optional): If False raise an error if the folder already exists.
            Defaults to True.
        parents (bool, optional): If True, also creates parent folders.
            Defaults to True.

    Returns:
        Path: Path to the created folder
    """
    if isinstance(path, str):
        path = Path(path)

    path.mkdir(exist_ok=exist_ok, parents=parents)

    return path



"""Nox configuration file."""
import nox


@nox.session
def tests(session):
    """Unit tests."""
    session.install(".[test]")
    session.run("pytest")


@nox.session
def bandit(session):
    """Bandit static security analysis."""
    session.install(".[test]")
    session.run("bandit", "-r", "src", "-c", "pyproject.toml")
